@@ -3,12 +3,18 @@
  * kind of integrals used for non-adiabatic molecular dynamics,
  * including the overlaps integrals between different geometries
  * And the dipoles and quadrupoles to compute absorption spectra.
- * This module is based on libint, Eigen and pybind11.
+ * This module is based on libint and Eigen.
  * Copyright (C) 2018-2022 the Netherlands eScience Center.
  */
+
+#define PY_SSIZE_T_CLEAN
+#define Py_LIMITED_API 0x03070000
+#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
+#include <Python.h>
+#include <numpy/arrayobject.h>
+
 #include "namd.hpp"
 
-namespace py = pybind11;
 using HighFive::Attribute;
 using HighFive::DataSet;
 using HighFive::File;
@@ -40,14 +46,11 @@ Matrix compute_integrals_couplings(const string &path_xyz_1,
                                    const string &basis_name);
 
 int main() {
-
-  string path_xyz = "../test/test_files/ethylene.xyz";
-  string path_hdf5 = "../test/test_files/ethylene.hdf5";
-  string basis_name = "DZVP-MOLOPT-SR-GTH";
-  string dataset_name = "ethylene/point_n";
-
-  auto xs =
-      compute_integrals_couplings(path_xyz, path_xyz, path_hdf5, basis_name);
+  const string path_xyz = "../test/test_files/ethylene.xyz";
+  const string path_hdf5 = "../test/test_files/ethylene.hdf5";
+  const string basis_name = "DZVP-MOLOPT-SR-GTH";
+  const string dataset_name = "ethylene/point_n";
+  auto xs = compute_integrals_couplings(path_xyz, path_xyz, path_hdf5, basis_name);
 }
 
 // OpenMP or multithread computations
@@ -67,13 +70,16 @@ template <typename Lambda> void parallel_do(Lambda &lambda) {
 #else // use C++11 threads
   std::vector<std::thread> threads;
   for (int thread_id = 0; thread_id != libint2::nthreads; ++thread_id) {
-    if (thread_id != nthreads - 1)
+    if (thread_id != nthreads - 1) {
       threads.push_back(std::thread(lambda, thread_id));
-    else
+    }
+    else {
       lambda(thread_id);
+    }
   } // threads_id
-  for (int thread_id = 0; thread_id < nthreads - 1; ++thread_id)
+  for (int thread_id = 0; thread_id < nthreads - 1; ++thread_id) {
     threads[thread_id].join();
+  }
 #endif
 }
 } // namespace libint2
@@ -89,44 +95,25 @@ void set_nthread() {
 #endif
 }
 
-/**
- * \brief Get the number of threads and the thread type as a 2-tuple
- */
-int get_thread_count() {
-  using libint2::nthreads;
-  return std::thread::hardware_concurrency();
-}
-
-/**
- * \brief Get the type of threads
- */
-string get_thread_type() {
-  string ret;
-
-  #if defined(_OPENMP)
-    ret = "OpenMP";
-  #else
-    ret = "C++11";
-  #endif
-  return ret;
-}
-
 /**
  * \brief Get the number of basis
  */
 size_t nbasis(const std::vector<libint2::Shell> &shells) {
-  return std::accumulate(cbegin(shells), cend(shells), 0,
-                         [](size_t acc, const libint2::Shell &shell) {
-                           return acc + shell.size();
-                         });
+  return std::accumulate(
+    shells.cbegin(),
+    shells.cend(),
+    0,
+    [](const size_t acc, const libint2::Shell &shell) { return acc + shell.size(); });
 }
+
 /**
  * \brief compute the maximum number of n-primitives
  */
 size_t max_nprim(const std::vector<libint2::Shell> &shells) {
   size_t n = 0;
-  for (const auto &shell : shells)
+  for (const auto &shell : shells) {
     n = std::max(shell.nprim(), n);
+  }
   return n;
 }
 
@@ -134,11 +121,12 @@ size_t max_nprim(const std::vector<libint2::Shell> &shells) {
  * \brief compute the maximum number of l-primitives
  */
 int max_l(const std::vector<libint2::Shell> &shells) {
-
   int l = 0;
-  for (const auto &shell : shells)
-    for (const auto &c : shell.contr)
+  for (const auto &shell : shells) {
+    for (const auto &c : shell.contr) {
       l = std::max(c.l, l);
+    }
+  }
   return l;
 }
 
@@ -179,10 +167,10 @@ Matrix compute_overlaps_for_couplings(const std::vector<Shell> &shells_1,
     engines[i] = engines[0];
   }
 
-  auto shell2bf = map_shell_to_basis_function(shells_1);
+  const auto shell2bf = map_shell_to_basis_function(shells_1);
 
   // Function to compute the integrals in parallel
-  auto compute = [&](int thread_id) {
+  auto compute = [&](const int thread_id) {
     // buf[0] points to the target shell set after every call  to
     // engine.compute()
     const auto &buf = engines[thread_id].results();
@@ -235,13 +223,13 @@ std::vector<Matrix> compute_multipoles(
   const auto n = nbasis(shells);
 
   std::vector<Matrix> result(nopers);
-  for (auto &r : result)
+  for (auto &r : result) {
     r = Matrix::Zero(n, n);
+  }
 
   // construct the multipole engine integrals engine
   std::vector<libint2::Engine> engines(nthreads);
-  engines[0] =
-      libint2::Engine(operator_type, max_nprim(shells), max_l(shells), 0);
+  engines[0] = libint2::Engine(operator_type, max_nprim(shells), max_l(shells), 0);
 
   // pass operator params to the engines
   engines[0].set_params(oparams);
@@ -249,7 +237,7 @@ std::vector<Matrix> compute_multipoles(
     engines[i] = engines[0];
   }
 
-  auto shell2bf = map_shell_to_basis_function(shells);
+  const auto shell2bf = map_shell_to_basis_function(shells);
 
   // Function to compute the integrals in parallel
   auto compute = [&](int thread_id) {
@@ -304,76 +292,71 @@ CP2K_Basis_Atom read_basis_from_hdf5(const string &path_file,
   libint2::svector<libint2::svector<double>> small_coef;
   libint2::svector<libint2::svector<CP2K_Contractions>> small_fmt;
 
-  try {
-    // Open an existing HDF5 File
-    const File file(path_file, File::ReadOnly);
+  // Open an existing HDF5 File
+  const File file(path_file, File::ReadOnly);
 
-    // build paths to the coefficients and exponents
-    const int n_elec = valence_electrons.at(symbol);
-    const string root = "cp2k/basis/" + symbol + "/" + basis + "-q" + std::to_string(n_elec);
+  // build paths to the coefficients and exponents
+  const int n_elec = valence_electrons.at(symbol);
+  const string root = "cp2k/basis/" + symbol + "/" + basis + "-q" + std::to_string(n_elec);
 
-    const Group group = file.getGroup(root);
-    const std::vector<string> dset_names = group.listObjectNames();
+  const Group group = file.getGroup(root);
+  const std::vector<string> dset_names = group.listObjectNames();
 
-    // Iterate over all exponent sets; for most basis sets there is
-    // only a single set of exponents, but there are exception such
-    // as BASIS_ADMM_MOLOPT
-    for (const auto &name : dset_names) {
-      std::vector<std::vector<double>> coefficients;
-      std::vector<double> exponents;
-      std::vector<int64_t> format;
+  // Iterate over all exponent sets; for most basis sets there is
+  // only a single set of exponents, but there are exception such
+  // as BASIS_ADMM_MOLOPT
+  for (const auto &name : dset_names) {
+    std::vector<std::vector<double>> coefficients;
+    std::vector<double> exponents;
+    std::vector<int64_t> format;
 
-      const string path_coefficients = root + "/" + name + "/coefficients";
-      const string path_exponents = root + "/" + name + "/exponents";
+    const string path_coefficients = root + "/" + name + "/coefficients";
+    const string path_exponents = root + "/" + name + "/exponents";
 
-      // Get the dataset
-      const DataSet dataset_cs = file.getDataSet(path_coefficients);
-      const DataSet dataset_es = file.getDataSet(path_exponents);
-      const Attribute attr = dataset_cs.getAttribute("basisFormat");
+    // Get the dataset
+    const DataSet dataset_cs = file.getDataSet(path_coefficients);
+    const DataSet dataset_es = file.getDataSet(path_exponents);
+    const Attribute attr = dataset_cs.getAttribute("basisFormat");
 
-      // extract data from the datasets
-      dataset_cs.read(coefficients);
-      dataset_es.read(exponents);
-      attr.read(format);
+    // extract data from the datasets
+    dataset_cs.read(coefficients);
+    dataset_es.read(exponents);
+    attr.read(format);
 
-      // Move data to small vectors and keep appending or extending them as
-      // iteration over `dset_names` continues
-      libint2::svector<double> small_exp_1d;
-      std::move(exponents.begin(), exponents.end(), std::back_inserter(small_exp_1d));
-      small_exp.push_back(small_exp_1d);
+    // Move data to small vectors and keep appending or extending them as
+    // iteration over `dset_names` continues
+    libint2::svector<double> small_exp_1d;
+    std::move(exponents.begin(), exponents.end(), std::back_inserter(small_exp_1d));
+    small_exp.push_back(small_exp_1d);
 
-      for (const auto &v : coefficients) {
-        libint2::svector<double> small_coef_1d;
-        std::move(v.begin(), v.end(), std::back_inserter(small_coef_1d));
-        small_coef.push_back(small_coef_1d);
-      }
-
-      // The CP2K basis format is defined by a vector of integers, for each atom.
-      // For example For the C atom and the Basis DZVP-MOLOPT-GTH the basis format
-      // is:
-      //  2 0 2 7 2 2 1
-      // where:
-      //   * 2 is the Principal quantum number
-      //   * 0 is the minimum angular momemtum l
-      //   * 2 is the maximum angular momentum l
-      //   * 7 is the number of total exponents
-      //   * 2 Contractions of the 0 + l_min (S) Gaussian Orbitals
-      //   * 2 Contractions of the 1 + l_min (P) Gaussian Orbitals
-      //   * 1 Contractions of the 2 + l_min (D) Gaussian Orbitals
-      //
-      // Note: Elements 4 and onwards define the number of contracted for each
-      // angular momentum quantum number (all prior elements are disgarded).
-      int l, i;
-      libint2::svector<CP2K_Contractions> small_fmt_1d;
-      for (i=4, l=format[1]; i != static_cast<int>(format.size()); i++, l++) {
-        int count = format[i];
-        small_fmt_1d.push_back({l, count});
-      }
-      small_fmt.push_back(small_fmt_1d);
+    for (const auto &v : coefficients) {
+      libint2::svector<double> small_coef_1d;
+      std::move(v.begin(), v.end(), std::back_inserter(small_coef_1d));
+      small_coef.push_back(small_coef_1d);
     }
-  } catch (HighFive::Exception &err) {
-    // catch and print any HDF5 error
-    std::cerr << err.what() << std::endl;
+
+    // The CP2K basis format is defined by a vector of integers, for each atom.
+    // For example For the C atom and the Basis DZVP-MOLOPT-GTH the basis format
+    // is:
+    //  2 0 2 7 2 2 1
+    // where:
+    //   * 2 is the Principal quantum number
+    //   * 0 is the minimum angular momemtum l
+    //   * 2 is the maximum angular momentum l
+    //   * 7 is the number of total exponents
+    //   * 2 Contractions of the 0 + l_min (S) Gaussian Orbitals
+    //   * 2 Contractions of the 1 + l_min (P) Gaussian Orbitals
+    //   * 1 Contractions of the 2 + l_min (D) Gaussian Orbitals
+    //
+    // Note: Elements 4 and onwards define the number of contracted for each
+    // angular momentum quantum number (all prior elements are disgarded).
+    int l, i;
+    libint2::svector<CP2K_Contractions> small_fmt_1d;
+    for (i=4, l=format[1]; i != static_cast<int>(format.size()); i++, l++) {
+      int count = format[i];
+      small_fmt_1d.push_back({l, count});
+    }
+    small_fmt.push_back(small_fmt_1d);
   }
   return CP2K_Basis_Atom{symbol, small_coef, small_exp, small_fmt};
 }
@@ -384,15 +367,24 @@ CP2K_Basis_Atom read_basis_from_hdf5(const string &path_file,
 std::vector<string> get_unique_symbols(const std::vector<Atom> &atoms) {
   // Return a set of unique symbols
   std::vector<int> elements;
-  std::transform(atoms.begin(), atoms.end(), std::back_inserter(elements),
-                 [](const Atom &at) { return at.atomic_number; });
+  std::transform(
+    atoms.begin(),
+    atoms.end(),
+    std::back_inserter(elements),
+    [](const Atom &at) { return at.atomic_number; }
+  );
+
   // Unique set of elements
-  std::unordered_set<int> set(elements.begin(), elements.end());
+  const std::unordered_set<int> set(elements.begin(), elements.end());
 
   // create a unique vector of symbols
   std::vector<string> symbols;
-  std::transform(set.cbegin(), set.cend(), std::back_inserter(symbols),
-                 [](int x) { return map_elements[x]; });
+  std::transform(
+    set.cbegin(),
+    set.cend(),
+    std::back_inserter(symbols),
+    [](const int x) { return map_elements.at(x); }
+  );
   return symbols;
 }
 
@@ -405,10 +397,9 @@ create_map_symbols_basis(const string &path_hdf5,
   std::unordered_map<string, CP2K_Basis_Atom> dict;
 
   // Select the unique atomic symbols
-  std::vector<string> symbols = get_unique_symbols(atoms);
+  const std::vector<string> symbols = get_unique_symbols(atoms);
   for (const auto &at : symbols)
     dict[at] = read_basis_from_hdf5(path_hdf5, at, basis);
-
   return dict;
 }
 
@@ -425,7 +416,7 @@ libint2::svector<Shell> create_shells_for_atom(const CP2K_Basis_Atom &data,
   for (int i = 0; i != static_cast<int>(data.exponents.size()); i++) {
     exponents = data.exponents[i];
     basis_format = data.basis_format[i];
-    for (auto contractions : basis_format) {
+    for (auto &contractions : basis_format) {
       for (int j = 0; j < contractions.count; j++) {
         shells.push_back({
           exponents,
@@ -447,17 +438,13 @@ std::vector<Shell> make_cp2k_basis(const std::vector<Atom> &atoms,
                                    const string &basis) {
   std::vector<Shell> shells;
 
-  // set of symbols
-  std::vector<string> symbols = get_unique_symbols(atoms);
-
   // Read basis set data from the HDF5
-  std::unordered_map<string, CP2K_Basis_Atom> dict =
+  const std::unordered_map<string, CP2K_Basis_Atom> dict =
       create_map_symbols_basis(path_hdf5, atoms, basis);
 
   for (const auto &atom : atoms) {
-
-    CP2K_Basis_Atom data = dict[map_elements[atom.atomic_number]];
-    auto xs = create_shells_for_atom(data, atom);
+    const CP2K_Basis_Atom data = dict.at(map_elements.at(atom.atomic_number));
+    const auto xs = create_shells_for_atom(data, atom);
     shells.insert(shells.end(), xs.begin(), xs.end());
   }
 
@@ -469,11 +456,11 @@ Matrix compute_integrals_couplings(const string &path_xyz_1,
                                    const string &basis_name) {
 
   set_nthread();
-  std::vector<Atom> mol_1 = read_xyz_from_file(path_xyz_1);
-  std::vector<Atom> mol_2 = read_xyz_from_file(path_xyz_2);
+  const std::vector<Atom> mol_1 = read_xyz_from_file(path_xyz_1);
+  const std::vector<Atom> mol_2 = read_xyz_from_file(path_xyz_2);
 
-  auto shells_1 = make_cp2k_basis(mol_1, path_hdf5, basis_name);
-  auto shells_2 = make_cp2k_basis(mol_2, path_hdf5, basis_name);
+  const auto shells_1 = make_cp2k_basis(mol_1, path_hdf5, basis_name);
+  const auto shells_2 = make_cp2k_basis(mol_2, path_hdf5, basis_name);
 
   // safe to use libint now
   libint2::initialize();
@@ -501,10 +488,12 @@ std::array<double, 3> calculate_center_of_mass(const std::vector<Atom> &atoms) {
     rs[1] += at.y * z, rs[2] += at.z * z;
   }
 
-  auto m =
-      std::accumulate(atoms.begin(), atoms.end(), 0, [](double acc, Atom at) {
-        return acc + double(at.atomic_number);
-      });
+  auto m = std::accumulate(
+    atoms.begin(),
+    atoms.end(),
+    0,
+    [](const double acc, const Atom &at) { return acc + double(at.atomic_number); }
+  );
   return {rs[0] / m, rs[1] / m, rs[2] / m};
 }
 
@@ -536,15 +525,15 @@ Matrix compute_integrals_multipole(const string &path_xyz,
                                    const string &basis_name,
                                    const string &multipole) {
   set_nthread();
-  std::vector<Atom> mol = read_xyz_from_file(path_xyz);
+  const std::vector<Atom> mol = read_xyz_from_file(path_xyz);
 
-  auto shells = make_cp2k_basis(mol, path_hdf5, basis_name);
+  const auto shells = make_cp2k_basis(mol, path_hdf5, basis_name);
 
   // safe to use libint now
   libint2::initialize();
 
   // compute Overlap integrals
-  auto matrices = select_multipole(mol, shells, multipole);
+  const auto matrices = select_multipole(mol, shells, multipole);
 
   // stop using libint2
   libint2::finalize();
@@ -559,19 +548,162 @@ Matrix compute_integrals_multipole(const string &path_xyz,
   return super_matrix;
 }
 
-PYBIND11_MODULE(compute_integrals, m) {
-  m.doc() = "Compute integrals using libint2 see: "
-            "https://github.com/evaleev/libint/wiki";
 
-  m.def("compute_integrals_couplings", &compute_integrals_couplings,
-        py::return_value_policy::reference_internal);
+/** \brief Convert python `str` instance into a C++ std::string **/
+int py_obj_to_string(PyObject *py_str, void *ptr) {
+  Py_ssize_t size;
+  char *str;
 
-  m.def("compute_integrals_multipole", &compute_integrals_multipole,
-        py::return_value_policy::reference_internal);
+  // PyUnicode_AsUTF8AndSize is part of the stable ABI ever since Python 3.10.
+  // For older versions we have to use a workaround by first converting the
+  // python str instance into a python bytes instance.
+#if defined(PyUnicode_AsUTF8AndSize)
+  str = PyUnicode_AsUTF8AndSize(py_str, &size);
+  if (str == nullptr) {
+    return 0;
+  }
+#else
+  PyObject *py_bytes = PyUnicode_AsEncodedString(py_str, "utf8", nullptr);
+  if (py_bytes == nullptr) {
+    return 0;
+  }
 
-  m.def("get_thread_count", &get_thread_count,
-        py::return_value_policy::reference_internal);
+  const int num = PyBytes_AsStringAndSize(py_bytes, &str, &size);
+  Py_DecRef(py_bytes);
+  if (num == -1) {
+    return 0;
+  }
+#endif
+  *(string *)ptr = std::string(str, size);
+  return 1;
+}
 
-  m.def("get_thread_type", &get_thread_type,
-        py::return_value_policy::reference_internal);
+
+/** \brief Convert (copy) an Eigen matrix into a numpy array **/
+PyObject *mat_to_npy_array(Matrix &mat) {
+  PyObject *npy_array;
+  PyObject *ret;
+  const npy_intp shape [2] = {mat.rows(), mat.cols()};
+
+  npy_array = PyArray_SimpleNewFromData(2, shape, NPY_DOUBLE, mat.data());
+  if (npy_array == nullptr) {
+    return nullptr;
+  }
+
+  // TODO: Figure out how to transfer ownership of `mat`s memory to the numpy array.
+  // As a stop-gap measure, simply create a copy to ensure the arrays' memory is
+  // fully managed by numpy.
+  ret = PyArray_SimpleNew(2, shape, NPY_DOUBLE);
+  int res = PyArray_CopyInto((PyArrayObject *)ret, (PyArrayObject *)npy_array);
+  Py_DecRef(npy_array);
+  return (res == -1) ? nullptr : ret;
+}
+
+
+/** \brief Python wrapper for compute_integrals_couplings */
+PyObject *py_compute_integrals_couplings(PyObject *self, PyObject *args) {
+  string path_xyz_1;
+  string path_xyz_2;
+  string path_hdf5;
+  string basis_name;
+  Matrix mat;
+
+  if (!PyArg_ParseTuple(args, "O&O&O&O&:compute_integrals_couplings",
+      py_obj_to_string, &path_xyz_1,
+      py_obj_to_string, &path_xyz_2,
+      py_obj_to_string, &path_hdf5,
+      py_obj_to_string, &basis_name)) {
+    return nullptr;
+  }
+
+  try {
+    mat = compute_integrals_couplings(path_xyz_1, path_xyz_2, path_hdf5, basis_name);
+  } catch (HighFive::Exception &err) {
+    PyErr_SetString(PyExc_RuntimeError, err.what());
+    return nullptr;
+  }
+  return mat_to_npy_array(mat);
+}
+
+
+/** \brief Python wrapper for compute_integrals_multipole */
+PyObject *py_compute_integrals_multipole(PyObject *self, PyObject *args) {
+  string path_xyz;
+  string path_hdf5;
+  string basis_name;
+  string multipole;
+  Matrix mat;
+
+  if (!PyArg_ParseTuple(args, "O&O&O&O&:compute_integrals_multipole",
+      py_obj_to_string, &path_xyz,
+      py_obj_to_string, &path_hdf5,
+      py_obj_to_string, &basis_name,
+      py_obj_to_string, &multipole)) {
+    return nullptr;
+  }
+
+  try {
+    mat = compute_integrals_multipole(path_xyz, path_hdf5, basis_name, multipole);
+  } catch (HighFive::Exception &err) {
+    PyErr_SetString(PyExc_RuntimeError, err.what());
+    return nullptr;
+  }
+  return mat_to_npy_array(mat);
+}
+
+
+/** \brief Get the number of threads */
+PyObject * py_get_thread_count(PyObject *self, PyObject *args) {
+  using libint2::nthreads;
+  return PyLong_FromLong(std::thread::hardware_concurrency());
+}
+
+
+/** \brief Get the type of threads */
+PyObject * py_get_thread_type(PyObject *self, PyObject *args) {
+  PyObject *ret;
+
+#if defined(_OPENMP)
+  ret = PyUnicode_FromString("OpenMP");
+#else
+  ret = PyUnicode_FromString("C++11");
+#endif
+  return ret;
+}
+
+
+PyMethodDef method_defs[] = {
+  {"compute_integrals_couplings", py_compute_integrals_couplings, METH_VARARGS,
+    "Compute the basis-set-specific overlap integrals for the molecule as provided in the xyz file."},
+  {"compute_integrals_multipole", py_compute_integrals_multipole, METH_VARARGS,
+    "Compute the given multipole."},
+  {"get_thread_count", py_get_thread_count, METH_NOARGS, "Get the number of threads."},
+  {"get_thread_type", py_get_thread_type, METH_NOARGS, "Get the type of threads."},
+  {nullptr}  // Sentinel
+};
+
+
+PyModuleDef py_module = {
+  PyModuleDef_HEAD_INIT,
+  "nanoqm.compute_integrals",                                              // m_name
+  "C++ extension module for computing integral overlaps and multipoles.",  // m_doc
+  -1,                                                                      // m_size
+  method_defs                                                              // m_methods
+};
+
+// Workaround to prevent C++ name mangling
+//
+// The `PyMODINIT_FUNC` macro should be able to take care of it, but it doesn't
+// seem to work on python <= 3.8.
+// Possibly related to the "-fvisibility=hidden" flag?
+
+#if defined(WIN32) || defined(_WIN32)
+  #define MODULE_EXPORT __declspec(dllexport)
+#else
+  #define MODULE_EXPORT __attribute__((visibility("default")))
+#endif
+
+extern "C" MODULE_EXPORT PyObject *PyInit_compute_integrals(void) {
+  import_array();
+  return PyModule_Create(&py_module);
 }
